#include <stdlib.h>
#include <string.h>

#include "block.h"

typedef struct _Block *Block;

#define BLOCK_POOL_MAGIC_NUMBER 0xb10c1000

typedef enum {
  BLOCK_TYPE_EDGE,
  BLOCK_TYPE_ALLOC,
  BLOCK_TYPE_FREE,
} block_type_t;

struct _Block {
  struct _Block *next;
  struct _Block *prev;
  block_type_t type;
  int size;
  char data[0];
};

struct _BlockList {
  struct _Block head;
  struct _Block tail;
};

struct _BlockPool {
  int magic;
  int size;
  struct _BlockList list;
};

static int BlockPool_Initialize(BlockPool pool, int size)
{
  Block block;

  memset(pool, 0, size);

  pool->magic = BLOCK_POOL_MAGIC_NUMBER;
  pool->size = size;

  block = (Block)(pool + 1);

  pool->list.head.type = BLOCK_TYPE_EDGE;
  pool->list.tail.type = BLOCK_TYPE_EDGE;
  pool->list.head.size = sizeof(pool->list.head);
  pool->list.tail.size = sizeof(pool->list.tail);

  pool->list.head.prev = NULL;
  pool->list.head.next = block;
  pool->list.tail.prev = block;
  pool->list.tail.next = NULL;

  block->prev = &pool->list.head;
  block->next = &pool->list.tail;
  block->type = BLOCK_TYPE_FREE;
  block->size = size - sizeof(*pool);

  return 0;
}

BlockPool BlockPool_Create(void *buffer, int size)
{
  BlockPool pool;

  pool = buffer;

  if (pool->magic != BLOCK_POOL_MAGIC_NUMBER) {
    BlockPool_Initialize(pool, size);
  }

  return pool;
}

BlockPool BlockPool_Destroy(BlockPool pool)
{
  memset(pool, 0, pool->size);
  return NULL;
}

void *BlockPool_Allocate(BlockPool pool, int size)
{
  Block block, new_block;
  int needsize, datasize;

  size = ((size + sizeof(int) - 1) / sizeof(int)) * sizeof(int);

  needsize = size + sizeof(*block);

  for (block = pool->list.head.next; block->type != BLOCK_TYPE_EDGE;
       block = block->next) {
    if (block->type != BLOCK_TYPE_FREE)
      continue;
    datasize = block->size - sizeof(*block);
    if (datasize < size)
      continue;

    block->type = BLOCK_TYPE_ALLOC;

    if (datasize > needsize) { /* ʬˤ */
      new_block = (Block)(((char *)block) + needsize);
      new_block->type = BLOCK_TYPE_FREE;
      new_block->size = block->size - needsize;
      new_block->prev = block;
      new_block->next = block->next;
      block->size = needsize;
      block->next = new_block;
    }

    return &block->data[0];
  }

  return NULL;
}

static int combine(Block block)
{
  Block next = block->next;
  if ((block->type == BLOCK_TYPE_FREE) && (next->type == BLOCK_TYPE_FREE)) {
    block->size += next->size;
    block->next = next->next;
    block->next->prev = block;
    memset(next, 0, sizeof(*block));
  }
  return 0;
}

int BlockPool_Free(BlockPool pool, void *p)
{
  Block block;

  block = (Block)((char *)p - sizeof(*block));
  memset(block->data, 0, block->size - sizeof(*block));

  block->type = BLOCK_TYPE_FREE;

  block = block->prev;
  combine(block);
  combine(block);

  return 0;
}
